#include "streams.h"
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>

#ifndef _WIN32
#define vsprintf_s vsnprintf
#endif

//--------------------- CStream ----------------
CStream::CStream() {}

CStream::~CStream() {}
	
int CStream::Size()
{
	int sz;
	int ps;
	ps = Seek(0,1);
	sz = Seek(0,2);
	Seek(ps,0);
	return sz;
}
	
int CStream::Position()
{
	int ps = Seek(0,1);
	return ps;
}


bool CStream::ReadBool()
{
	bool val = false;
	Read(&val,1);
	return val;
}
	
unsigned char CStream::ReadByte()
{
	unsigned char val = 0;
	Read(&val,1);
	return val;
}
	
unsigned short CStream::ReadShort()
{
	unsigned short val = 0;
	Read(&val,2);
	return val;
}

int CStream::ReadLong()
{
	int val = 0;
	Read(&val,4);
	return val;
}

float CStream::ReadFloat()
{
	float val = 0.0f;
	Read(&val,4);
	return val;
}

double CStream::ReadDouble()
{
	double val = 0.0f;
	Read(&val,8);
	return val;
}

unsigned long long CStream::ReadUInt64()
{
	unsigned long long val = 0LL;
	Read(&val,8);
	return val;
}

bool CStream::WriteUInt64(unsigned long long val)
{
	return Write(&val,8)==8;
}


char *CStream::ReadText(char *buffer, size_t bufferSize)
{
	buffer[0] = 0;

	char c;
	do
	{
		 c = (char)ReadByte();
	} while(c==32);

	if (!c) return NULL;

	size_t w = 0;
	do
	{
		if (w < bufferSize-1) buffer[w++] = c;
		c = (char)ReadByte();
	} while(c!=0 && c!=32);
	buffer[w] = 0;
	return buffer;
}

int CStream::ReadTextInt(int def)
{
	char buf[32];
	if (!ReadText(buf, sizeof(buf))) return def;
	errno = 0;
	int res = atoi(buf);
	if (errno) res = def;
	return res;
}

double CStream::ReadTextFloat(double def)
{
	char buf[64];
	if (!ReadText(buf, sizeof(buf))) return def;
	errno = 0;
	double res = atof(buf);
	if (errno) res = def;
	return res;
}






char* CStream::ReadAnsiString()
{
	int savePos = Position();
	char c;
	int newPos = savePos;
	int sz = Size();
	do
	{
		Read(&c,1);
		newPos++;
	} while (c && newPos < sz);

	if (newPos <= savePos + 1) return NULL;

	Seek(savePos,0);
	
	int len = newPos - savePos;
	char *ansiString = (char*)malloc(len);
	Read(ansiString, len);
	ansiString[len-1] = 0;

	return ansiString;
}


bool CStream::WriteBool(bool val)
{
	return Write(&val,1)==1;
}
	
bool CStream::WriteByte(unsigned char val)
{
	return Write(&val,1)==1;
}
	
bool CStream::WriteShort(unsigned short val)
{
	return Write(&val,2)==2;
}
	
bool CStream::WriteLong(int val)
{
	return Write(&val,4)==4;
}

bool CStream::WriteFloat(float val)
{
	return Write(&val,4)==4;
}

bool CStream::WriteDouble(double val)
{
	return Write(&val,8)==8;
}

	
bool CStream::WriteString(const char * val)
{
	if (!val)
	{
		char z=0;
		if (Write(&z, sizeof(z))!=(int)sizeof(z)) return false;
	}
	else
	{
		int n = (int)(strlen(val)+1);
		if (Write((void*)val, n)!=n) return false;
	}
	return true;
}

bool CStream::WriteText(const char *text)
{
	int len = (int)strlen(text);
	return Write(text, len)==len;
}

bool CStream::Format(const char *format,...)
{
	va_list va;
	va_start(va, format);
	char buffer[8000];
	vsprintf_s(buffer, sizeof(buffer)-1, format, va);
	va_end(va);
	return WriteText(buffer);
}

	
bool CStream::CopyFrom(CStream *source, int count)
{
	const int maxBlockSize = 0x10000;
	void *buffer;
	int last, ps, sz;


	sz = source->Size();
	ps = source->Position();

	if (count<0 || count > sz - ps)
	{
		count = sz - ps;
	}
	if (count > 0)
	{
		buffer = malloc(maxBlockSize);
		if (!buffer) return false;
		while (count>0)
		{
			last = count;
			if (last>maxBlockSize) last = maxBlockSize;
			int n = source->Read(buffer, last);
			if (n!=last) { free(buffer); return false; }
			if (Write(buffer, n)!=n) { free(buffer); return false; }
			count-=last;
		}
		free(buffer);
	}
	return true;
}


//-------------- CFileStream --------------------
CFileStream::CFileStream(const TCHAR *fileName, const TCHAR *mode) : CStream()
{
	#ifdef _WIN32
	if (_tfopen_s(&f, fileName, mode)!=0) f = NULL;
	#else
	f = fopen(fileName, mode);
	#endif
}
	
CFileStream::~CFileStream()
{
	if (f) fclose(f);
}

int CFileStream::Write(const void *mem, int count)
{
	if (!f) return -1;
	return (int)fwrite(mem, 1, count, f);
}
	
int CFileStream::Read(void *mem, int count)
{
	if (!f) return -1;
	return (int)fread(mem, 1, count, f);
}

int CFileStream::Seek(int offset, int origin)
{
	if (!f) return -1;
	fseek(f, offset, origin);
	return (int)ftell(f);
}

void CFileStream::Close()
{
	if (!f) return;
	fclose(f);
	f = NULL;
}


//--------------------------- CMemoryStream ------------------------//
CMemoryStream::CMemoryStream():CStream()
{
	memory = NULL;
	size = 0;
	position = 0;
	blockSize = 1024;
}
	
CMemoryStream::~CMemoryStream()
{
	free(memory);
}

void CMemoryStream::Cut(int pos)
{
	int last = size - pos;
	if (last > 0)
	{
		memmove(memory, memory + pos, last);
	}
	SetSize(last);
	position = 0;
}

void CMemoryStream::Cut()
{
	Cut(position);
}

int CMemoryStream::Find(const char *str, int len)
{
	return Find(0, str, len);
}

int CMemoryStream::FindEnd(int startPos, const char *str, int len)
{
	int p = Find(startPos, str, len);
	if (p >=0)
	{
		if (len < 0) len = (int)strlen(str);
		p+=len;
	}
	return p;
}

int CMemoryStream::FindEnd(const char *str, int len)
{
	return FindEnd(0, str, len);
}


int CMemoryStream::Find(int startPos, const char *str, int len)
{
	if (!str || !str[0]) return -1;
	if (len < 0) len = (int)strlen(str);

	for(int i=startPos; i <= size - len; i++)
	{
		if (memory[i] == str[0])
		{
			bool ok = true;
			for(int j=1; j < len; j++)
			{
				if (memory[i+j] != str[j])
				{
					ok = false;
					break;
				}
			}
			if (ok)
			{
				return i;
			}
		}
	}
	return -1;
}

bool CMemoryStream::Insert(int pos, const char *str, int len)
{
	if (pos < 0 || pos > size) pos = size;
	if (len < 0) len = (int)strlen(str);

	int oldSize = size;
	unsigned char *mem = memory;

	memory = NULL;
	size = 0;
	position = 0;

	if (!SetSize(oldSize + len))
	{
		memory = mem;
		size = oldSize;
		return false;
	}

	if (pos > 0) memcpy(memory, mem, pos);
	memcpy(memory+pos, str, len);
	if (oldSize - pos > 0) memcpy(memory+pos+len, mem + pos, oldSize - pos);
	free(mem);

	return true;
}

void CMemoryStream::Delete(int startPos, int len)
{
	if (len < 0)
	{
		if (startPos < 0 || startPos > size) return;
		SetSize(startPos);
		return;
	}
	int last = size - startPos + len;
	if (last > 0) memmove(memory + startPos, memory + startPos + len, last);
	SetSize(size - len);
}

	
int CMemoryStream::Write(const void *mem, int count)
{
	int end;
	if (count <=0) return 0;
	end = position + count;
	if (end > size)
	{
		if (!SetSize(end)) return 0;
	}
	memcpy(memory + position, mem, (size_t)count);
	position+=count;
	return count;
}
	
int CMemoryStream::Read(void *mem, int count)
{
	if (count <= 0) return 0;
	if (count + position > size) count = size - position;
	if (count <=0) return 0;
	
	memcpy(mem, memory + position, (size_t)count);
	position+=count;
	return count;
}

int CMemoryStream::Seek(int offset, int origin)
{
	switch(origin)
	{
		case 0:	position = offset; break;
		case 1: position+= offset; break;
		case 2: position = size + offset; break;
	}
	if (position > size) position = size;
	else
	if (position < 0) position = 0;
	return position;
}
	
bool CMemoryStream::SetSize(int newSize)
{
	int newSize2;
	int size2;
	int blk;
	int iblk;

	blk = blockSize-1;
	iblk = ~blk;

	if (newSize<0) newSize = 0;

	size2 = (size + blk) & iblk;
	newSize2 = (newSize + blk) & iblk;

	if (size2 != newSize2)
	{
		if (newSize2 > 0)
		{
			unsigned char *newMem = (unsigned char*)realloc(memory, newSize2);
			if (!newMem) return false;
			memory = newMem;
		}
		else
		{
			free(memory);
			memory = NULL;
		}
	}
	size = newSize;
	if (position > size) position = size;
	return true;
}

bool CMemoryStream::SaveToFile(const TCHAR *fileName)
{
	CFileStream fstream(fileName, _T("wb"));
	if (!fstream.f) return false;

	position = 0;
	return fstream.CopyFrom( this, size );
}

bool CMemoryStream::LoadFromFile(const TCHAR *fileName)
{
	CFileStream fstream(fileName, _T("rb"));
	if (!fstream.f) return false;

	bool res = CopyFrom(&fstream, fstream.Size());
	position = 0;
	return res;
}


